from PyQt5 import QtWidgets
from PyQt5 import QtCore
from PyQt5 import QtGui
from popplerqt5 import Poppler

from utils import debug

class StyledItemDelegate(QtWidgets.QStyledItemDelegate):
    """
    style definition for items in tableview
    """
    def initStyleOption(self, option, index):
        super(StyledItemDelegate, self).initStyleOption(option, index)
        # option.decorationPosition = QtWidgets.QStyleOptionViewItem.Top
        # option.displayAlignment = QtCore.Qt.AlignHCenter | QtCore.Qt.AlignTop

class AnnotTableView(QtWidgets.QTableView):
    # rendered = QtCore.pyqtSignal(int, QtGui.QImage)
    # textFound = QtCore.pyqtSignal(int, list)

    resizeRequested = QtCore.pyqtSignal(int, int)

    def __init__(self, parent):
        super(AnnotTableView, self).__init__(parent)
        # self.horizontalHeader().setStretchLastSection(True)
        self.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Stretch)
        # self.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)

        # header.setSectionResizeMode(1, QtWidgets.QHeaderView.ResizeToContents)
        # header.setSectionResizeMode(2, QtWidgets.QHeaderView.ResizeToContents)

        self.horizontalHeader().hide()
        self.verticalHeader().hide()
        
        self.setItemDelegate(StyledItemDelegate(self)) # customed style

        self.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers) # make items readonly

        # self.horizontalHeader().setSizeAdjustPolicy(QtWidgets.Q)
        
    # def viewOptions(self):
    #     # https://stackoverflow.com/questions/56438540/align-pixmap-in-columns-center-with-qtreeview-and-qabstractitemmodel
    #     # align the icons of all items centerally
    #     option = super().viewOptions()
    #     option.decorationAlignment = (QtCore.Qt.AlignHCenter | QtCore.Qt.AlignCenter)
    #     option.decorationPosition = QtWidgets.QStyleOptionViewItem.Top
    #     return option

    def resizeEvent(self, ev):
        debug("resizeEvent in AnnotTableView")
        print(self.width(), self.height())
        QtWidgets.QTableView.resizeEvent(self, ev) # call parent's event handler
        self.resizeRequested.emit(self.width(), self.height())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys, os

from subprocess import Popen

from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

# add parent directory to path
currentDir = os.path.dirname(os.path.abspath(__file__))
parentDir = os.path.dirname(currentDir)
# debug(currentDir)
# debug(parentDir)
sys.path.append(currentDir)
# sys.path.append(parentDir)
sys.path.append(currentDir + os.path.sep + 'resources')

from utils import debug

from resources.ui_main import Ui_window

from document import DocumentView
from library import LibraryView

from lineedit import PageNoLineEdit, FindLineEdit

from dialogs import ExportToImageDialog, DocInfoDialog

from docgraphicsview import DocGraphicsView

HOMEDIR = os.path.expanduser("~")

class MainWindow(QtWidgets.QMainWindow, Ui_window):
    def __init__(self, screens):
        super(MainWindow, self).__init__() # Call the inherited classes __init__ method
        self.setupUi(self)

        self.screen_dpi = screens[0].logicalDotsPerInch()
        
        self.libWidget = LibraryView(self.centraltabwidget, self.screen_dpi)
        self.libWidget.preview_graphicsview.viewportChanged.connect(self.setPageInfoOnToolbar)
        self.libWidget.preview_graphicsview.zoomRatioChanged.connect(self.onZoomRatioChanged)
        self.libWidget.preview_graphicsview.viewColumnChanged.connect(self.onViewColumnChanged)
        self.libWidget.preview_graphicsview.emptyLeadingPageChanged.connect(self.onEmptyLeadingPageChanged)
        self.centraltabwidget.addTab(self.libWidget, "My Library")
        
        self.centraltabwidget.setTabsClosable(True)

        # https://stackoverflow.com/questions/2616483/close-button-only-for-some-tabs-in-qt
        # make the first tab unclosable
        self.centraltabwidget.tabBar().tabButton(0, QtWidgets.QTabBar.RightSide).resize(0,0)

        self.centraltabwidget.currentChanged.connect(self.onTabChanged)
        self.centraltabwidget.tabCloseRequested.connect(self.onTabClose)
        self.centraltabwidget.setStyleSheet("QTabWidget::pane {margin: 0 0 0 0}")

        # self.dockSearch.hide()
        # self.dockWidget.hide()
        # self.dockWidget.setMinimumWidth(310)
        # self.findTextEdit.setFocusPolicy(QtCore.Qt.StrongFocus)

        # resizing pages requires some time to take effect
        # self.resize_page_timer = QtCore.QTimer(self)
        # self.resize_page_timer.setSingleShot(True)
        # self.resize_page_timer.timeout.connect(self.onWindowResize)
        # Add shortcut actions
        # self.gotoPageAction = QtWidgets.QAction(QIcon(":/goto.png"), "GoTo Page", self)
        # self.gotoPageAction.triggered.connect(self.gotoPage)
        # self.copyTextAction = QtWidgets.QAction(QIcon(":/copy.png"), "Copy Text", self)
        # self.copyTextAction.setCheckable(True)
        # self.copyTextAction.triggered.connect(self.toggleCopyText)
        # self.findTextAction = QtWidgets.QAction(QIcon(":/search.png"), "Find Text", self)
        # self.findTextAction.setShortcut('Ctrl+F')
        # self.findTextAction.triggered.connect(self.dockSearch.show)

        # connect menu actions signals
        self.openFileAction.triggered.connect(self.openFile)
        # self.printAction.triggered.connect(self.printFile)
        # self.quitAction.triggered.connect(self.close)
        # self.toPSAction.triggered.connect(self.exportToPS)
        # self.pageToImageAction.triggered.connect(self.exportPageToImage)
        # self.docInfoAction.triggered.connect(self.docInfo)
        self.zoominAction.triggered.connect(self.zoomIn)
        self.zoomoutAction.triggered.connect(self.zoomOut)
        self.zoomFitwidthAction.triggered.connect(self.zoomFitWidth)
        
        self.zoomFitwidthAction.setChecked(True)

        self.actionViewInOneColumn.triggered.connect(self.onViewInOneColumn)
        self.actionViewInTwoColumns.triggered.connect(self.onViewInTwoColumns)
        self.actionPrecedingEmptyPage.triggered.connect(self.onPrecedingEmptyPage)

        # default is one column
        self.actionViewInOneColumn.setChecked(True)

        # self.undoJumpAction.triggered.connect(self.undoJump)
        self.prevPageAction.triggered.connect(self.goPrevPage)
        self.nextPageAction.triggered.connect(self.goNextPage)
        self.actionFind.triggered.connect(self.onActionFind)
        # self.lastPageAction.triggered.connect(self.goLastPage)

        self.pageNoEdit = PageNoLineEdit(self)
        self.pageNoEdit.returnPressed.connect(self.gotoPage)

        self.findTxtEdit = FindLineEdit(self)

        self.pageNoLabel = QtWidgets.QLabel(self)
        self.pageNoLabel.setFrameShape(QtWidgets.QFrame.StyledPanel)
        self.pageNoLabel.setText('of')

        # self.viewColumnNumberCombo = QtWidgets.QComboBox(self)
        # self.viewColumnNumberCombo.activated.connect(self.setViewColumnNumber)
        # self.viewColumnNumberCombo.addItems(["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"])
        # self.zoom_levels = [0, 75, 90, 100, 110 , 121, 133, 146, 175, 200, 400, 800, 1600]

        # Add toolbar actions
        self.toolBar.addAction(self.undoJumpAction)
        self.toolBar.addAction(self.redoJumpAction)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.openFileAction)
        self.toolBar.addAction(self.printAction)
        # self.toolBar.addAction(self.docInfoAction)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.zoomoutAction)
        self.toolBar.addAction(self.zoominAction)
        self.toolBar.addAction(self.zoomFitwidthAction)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionViewInOneColumn)
        self.toolBar.addAction(self.actionViewInTwoColumns)
        self.toolBar.addAction(self.actionPrecedingEmptyPage)
        self.toolBar.addSeparator()
        # self.toolBar.addAction(self.firstPageAction)
        self.toolBar.addAction(self.prevPageAction)
        self.toolBar.addWidget(self.pageNoEdit)
        self.toolBar.addWidget(self.pageNoLabel)
        self.toolBar.addAction(self.pageCountAction)
        self.toolBar.addAction(self.nextPageAction)
        # self.toolBar.addAction(self.lastPageAction)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.actionFind)
        self.toolBar.addWidget(self.findTxtEdit)
        # self.toolBar.addAction(self.redoModificationAction)
        spacer = QtWidgets.QWidget(self)
        spacer.setSizePolicy(1|2|4,1|4)
        self.toolBar.addWidget(spacer)
        self.toolBar.addSeparator()
        # self.toolBar.addAction(self.quitAction)

        # # Add widgets
        # self.statusbar = QLabel(self)
        # self.statusbar.setStyleSheet("QLabel { font-size: 12px; border-radius: 2px; padding: 2px; background: palette(highlight); color: palette(highlighted-text); }")
        # self.statusbar.setMaximumHeight(16)
        # self.statusbar.hide()

        # Import settings
        desktop = QtWidgets.QApplication.desktop()
        self.settings = QtCore.QSettings("kuafu", "main", self)
        self.recent_files = self.settings.value("RecentFiles", [])
        self.history_filenames = self.settings.value("HistoryFileNameList", [])
        self.history_page_no = self.settings.value("HistoryPageNoList", [])
        self.offset_x = int(self.settings.value("OffsetX", 4))
        self.offset_y = int(self.settings.value("OffsetY", 26))
        self.available_area = [desktop.availableGeometry().width(), desktop.availableGeometry().height()]
        # self.zoomLevelCombo.setCurrentIndex(int(self.settings.value("ZoomLevel", 2)))

        # Connect Signals
        # self.findTextEdit.returnPressed.connect(self.findNext)
        # self.findNextButton.clicked.connect(self.findNext)
        # self.findBackButton.clicked.connect(self.findBack)
        # self.findCloseButton.clicked.connect(self.dockSearch.hide)
        # self.dockSearch.visibilityChanged.connect(self.toggleFindMode)

        self.recent_files_actions = []
        self.addRecentFiles()

        # Show Window
        width = int(self.settings.value("WindowWidth", 1040))
        height = int(self.settings.value("WindowHeight", 717))

        self.resize(width, height)
        self.show()

    def showStatus(self, msg):
        self.statusBar.showMessage(msg, msecs=5000)
        # if url=="":
        #     self.statusbar.hide()
        #     return
        # self.statusbar.setText(url)
        # self.statusbar.adjustSize()
        # self.statusbar.move(0, self.height()-self.statusbar.height())
        # self.statusbar.show()

    def onActionFind(self):
        self.findTxtEdit.setFocus()

    def onViewInOneColumn(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        widget.setColumnNumber(1)

    def onViewInTwoColumns(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        widget.setColumnNumber(2)

    def onPrecedingEmptyPage(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        if self.actionPrecedingEmptyPage.isChecked():
            widget.setPrecedingEmptyPage(1)
        else:
            widget.setPrecedingEmptyPage(0)

    def zoomIn(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        self.zoomFitwidthAction.setChecked(False)
        widget.zoomIn()

    def zoomOut(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        self.zoomFitwidthAction.setChecked(False)
        widget.zoomOut()

    def zoomFitWidth(self):
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        # if isinstance(widget, DocumentView):
        self.zoomFitwidthAction.setChecked(True)
        widget.zoomFitWidth()

    def goPrevPage(self):
        debug("goPrevPage")
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        if isinstance(widget, DocumentView):
            widget.goPrevPage()

    def goNextPage(self):
        debug("goNextPage")
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        if isinstance(widget, DocumentView):
            widget.goNextPage()

    def gotoPage(self):
        debug("gotoPage")
        page_no = int(self.pageNoEdit.text())
        cIdx = self.centraltabwidget.currentIndex()
        widget = self.centraltabwidget.widget(cIdx)
        if isinstance(widget, DocumentView):
            widget.gotoPage(page_no - 1)
            self.pageNoEdit.clearFocus()

    def onZoomRatioChanged(self, zoomRatio):
        if zoomRatio == 0:
            self.showStatus("Zoom fitted to width")
        else:
            self.zoomFitwidthAction.setChecked(False)
            self.showStatus("Zoom to %d%%" % int(zoomRatio * 100))
    
    def onViewColumnChanged(self, viewColumn):
        if viewColumn == 1:
            self.actionViewInOneColumn.setChecked(True)
            self.actionViewInTwoColumns.setChecked(False)
        elif viewColumn == 2:
            self.actionViewInOneColumn.setChecked(False)
            self.actionViewInTwoColumns.setChecked(True)
        else:
            assert(0)

    def onEmptyLeadingPageChanged(self, emptyPages):
        if emptyPages == 0:
            self.actionPrecedingEmptyPage.setChecked(False)
        elif emptyPages == 1:
            self.actionPrecedingEmptyPage.setChecked(True)
        else:
            assert(0)

    def setPageInfoOnToolbar(self, filename, page_counts, visible_regions):
        if len(visible_regions) == 0:
            return
        curent_page_idx = next(iter(visible_regions)) # fetch first key
        # 
        self.pageNoEdit.setMaxLength(len(str(page_counts)))
        self.pageNoEdit.setText("%d" % (curent_page_idx +1))
        self.pageCountAction.setText("%d" % page_counts)

    def addRecentFiles(self):
        self.recent_files_actions[:] = [] # pythonic way to clear list
        self.menuRecentFiles.clear()
        for each in self.recent_files:
            name = elideMiddle(os.path.basename(each), 60)
            action = self.menuRecentFiles.addAction(name, self.openRecentFile)
            self.recent_files_actions.append(action)
        self.menuRecentFiles.addSeparator()
        self.menuRecentFiles.addAction(QtGui.QIcon(':/edit-clear.png'), 'Clear Recents', self.clearRecents)

    def openRecentFile(self):
        action = self.sender()
        index = self.recent_files_actions.index(action)
        self.loadPDFfile(self.recent_files[index])

    def clearRecents(self):
        self.recent_files_actions[:] = []
        self.menuRecentFiles.clear()
        self.recent_files[:] = []

    def removeOldDoc(self):
        # Save current page number
        self.saveFileData()
        # Remove old document
        for i in range(len(self.pages)):
            self.verticalLayout.removeWidget(self.pages[-1])
        for i in range(len(self.pages)):
            self.pages.pop().deleteLater()
        self.frame.deleteLater()
        self.jumped_from = None
        #self.addRecentFiles()

    def loadPDFfile(self, filename):
        """ Loads pdf document in all threads """
        filename = os.path.expanduser(filename)
        
        # widget = DocumentView(self.centraltabwidget, filename, self.screen_dpi)
        # widget.showStatusRequested.connect(self.showStatus)
        # self.centraltabwidget.addTab(widget, filename)
        # self.centraltabwidget.setCurrentWidget(widget)

        # self.annotview_model.removeRows(0, self.annotview_model.rowCount())
        self.libWidget.preview_graphicsview.setDocument(filename, self.screen_dpi)
        

    def openFile(self):
        filename, sel_filter = QtWidgets.QFileDialog.getOpenFileName(self,
                                      "Select Document to Open", "",
                                      "Portable Document Format (*.pdf);;All Files (*)" )
        if filename != "":
            self.loadPDFfile(filename)

    def printFile(self):
        Popen(["quikprint", self.filename])

    def exportToPS(self):
        width = self.doc.page(self.current_page-1).pageSizeF().width()
        height = self.doc.page(self.current_page-1).pageSizeF().height()
        filename, sel_filter = QtWidgets.QFileDialog.getSaveFileName(self, "Select File to Save",
                                       os.path.splitext(self.filename)[0]+'.ps',
                                      "Adobe Postscript Format (*.ps)" )
        if filename == '' : return
        conv = self.doc.psConverter()
        conv.setPaperWidth(width)
        conv.setPaperHeight(height)
        conv.setOutputFileName(filename)
        conv.setPageList([i+1 for i in range(self.pages_count)])
        ok = conv.convert()
        if ok:
            QMessageBox.information(self, "Successful !","File has been successfully exported")
        else:
            QMessageBox.warning(self, "Failed !","Failed to export to Postscript")

    def exportPageToImage(self):
        dialog = ExportToImageDialog(self.current_page, self.pages_count, self)
        if dialog.exec_() == QDialog.Accepted:
            try:
                dpi = int(dialog.dpiEdit.text())
                page_no = dialog.pageNoSpin.value()
                filename = os.path.splitext(self.filename)[0]+'-'+str(page_no)+'.jpg'
                page = self.doc.page(page_no-1)
                if not page : return
                img = page.renderToImage(dpi, dpi)
                img.save(filename)
                QMessageBox.information(self, "Successful !","Page has been successfully exported")
            except:
                QMessageBox.warning(self, "Failed !","Failed to export to Image")

    def onTabChanged(self, tabIdx):
        debug("tab changed")
        widget = self.centraltabwidget.widget(tabIdx)
        if isinstance(widget, DocumentView):
            currentPage = widget.current_page
            currentPageCount = widget.pages_count
            self.setPageInfoOnToolbar(currentPage, currentPageCount)
            widget.pagePositionChanged.connect(self.setPageInfoOnToolbar)

    def saveFileData(self, filename, current_page):
        if filename != '':
            filename = collapseUser(filename)
            if filename in self.history_filenames:
                index = self.history_filenames.index(filename)
                self.history_page_no[index] = current_page
            else:
                self.history_filenames.insert(0, filename)
                self.history_page_no.insert(0, current_page)
            if filename in self.recent_files:
                self.recent_files.remove(filename)
            self.recent_files.insert(0, filename)

    def onTabClose(self, tabIdx):
        widget = self.centraltabwidget.widget(tabIdx)
        # save information if necessary
        if isinstance(widget, DocumentView):
            filename = widget.filename
            current_page = widget.current_page
            self.saveFileData(filename, current_page)
            # 
            self.settings.setValue("OffsetX", self.geometry().x()-self.x())
            self.settings.setValue("OffsetY", self.geometry().y()-self.y())
            # self.settings.setValue("ZoomLevel", self.zoom_level)
            self.settings.setValue("HistoryFileNameList", self.history_filenames[:100])
            self.settings.setValue("HistoryPageNoList", self.history_page_no[:100])
            self.settings.setValue("RecentFiles", self.recent_files[:10]) 
        # remove
        widget.close()
        self.centraltabwidget.removeTab(tabIdx)

    def onAppQuit(self):
        debug("OnAppQuit")
        """ Close running threads """
        tabCnt = self.centraltabwidget.count()
        for i in range(tabCnt):
            self.centraltabwidget.widget(i).close()

def wait(millisec):
    loop = QtCore.QEventLoop()
    QtCore.QTimer.singleShot(millisec, loop.quit)
    loop.exec_()

def elideMiddle(text, length):
    if len(text) <= length: return text
    return text[:length//2] + '...' + text[len(text)-length+length//2:]

def collapseUser(path):
    ''' converts /home/user/file.ext to ~/file.ext '''
    if path.startswith(HOMEDIR):
        return path.replace(HOMEDIR, '~', 1)
    return path

def main():
    app = QtWidgets.QApplication(sys.argv)
    win = MainWindow(app.screens())
    if len(sys.argv)>1 and os.path.exists(os.path.abspath(sys.argv[-1])):
        win.loadPDFfile(os.path.abspath(sys.argv[-1]))
    app.aboutToQuit.connect(win.onAppQuit)
    sys.exit(app.exec_())

if __name__ == "__main__":
    main()



 
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from popplerqt5 import Poppler
from pdfworker import PdfRender
import cv2
import numpy as np

from utils import debug
import math

def cvImg2QImg(cvImg):
    height, width, channel = cvImg.shape
    assert(channel == 4)
    bytesPerLine = channel * width
    qImg = QtGui.QImage(cvImg.data, width, height, bytesPerLine, QtGui.QImage.Format_RGBA8888)
    return qImg

def QImg2cvImg(qImg):
    qImg = qImg.convertToFormat(QtGui.QImage.Format_RGBA8888)
    width = qImg.width()
    height = qImg.height()
    ptr = qImg.bits() # enforce deep copy
    ptr.setsize(qImg.byteCount())
    arr = np.array(ptr).reshape(height, width, 4)
    return arr

class PageGraphicsItem(QtWidgets.QGraphicsRectItem):
    def __init__(self, pageIdx, parent=None):
        super(PageGraphicsItem, self).__init__(parent)

        self.pageIdx = pageIdx 

        self.setRect(0,0,0,0)
        self.setPos(0, 0)

        pen = QtGui.QPen(QtCore.Qt.NoPen) # remove rect border
        # pen = QtGui.QPen(QtCore.Qt.black)
        # pen.setWidth(1)
        self.setPen(pen)
        self.setBrush(QtCore.Qt.white) # fill white color

        self.borderItem = QtWidgets.QGraphicsRectItem()
        self.borderItem.setRect(0, 0, 0, 0)
        pen = QtGui.QPen(QtCore.Qt.gray)
        pen.setWidth(1)
        self.borderItem.setPen(pen)
        self.borderItem.setParentItem(self)

        # pixmaps
        self.patch_basesize = 800
        self.patch_rects = []
        self.patch_row_num = 0
        self.patch_col_num = 0
        
        # for smoothing transitions
        self.cached_pixmaps = {}
        self.current_items = {}

        # mask
        self.maskItem = QtWidgets.QGraphicsRectItem()
        self.maskItem.setRect(0, 0, 0, 0)

        pen = QtGui.QPen(QtCore.Qt.yellow)
        pen.setWidth(1)
        self.maskItem.setPen(pen)
        self.maskItem.setBrush(QtGui.QBrush(QtGui.QColor(0, 0, 0, 100))) # fill color

        self.maskItem.setParentItem(self)

        # 
        self.setZValue(0)
        self.borderItem.setZValue(3)
        self.maskItem.setZValue(4)
        # 0: self, 1: transient, 2: current, 3: border, 4: mask

    def initialize(self, x, y, width, height):
        self.patch_rects = self.compute_patch_rects(width, height)

        # may be called many times before the first addPixmap()
        current_width = self.rect().width()
        ratio_thistime = width / current_width if current_width > 0 else 1.0

        self.setPos(x, y)
        self.setRect(0, 0, width, height)
        self.borderItem.setRect(0, 0, width, height)

        # set mask item accordingly
        mask_rect = self.maskItem.rect()
        self.maskItem.setRect(
            mask_rect.x() * ratio_thistime, 
            mask_rect.y() * ratio_thistime, 
            mask_rect.width() * ratio_thistime, 
            mask_rect.height() * ratio_thistime
        )

        # remove all transient items
        for pid in self.cached_pixmaps:
            associated_item = self.cached_pixmaps[pid]['item']
            if associated_item:
                associated_item.setParentItem(None)
                self.cached_pixmaps[pid]['item'] = None

        # remove all current items (move their pixmaps to cache)
        for pid in self.current_items:
            self.current_items[pid]['item'].setParentItem(None)
            self.cached_pixmaps[pid] = self.current_items[pid]
            self.cached_pixmaps[pid]['item'] = None
        self.current_items = {}

        # update the ratio in cached pixmaps
        for pid in self.cached_pixmaps:
            self.cached_pixmaps[pid]['ratio'] *= ratio_thistime

        self.setVisible(True)

    def setBorderHighlight(self, active):
        if active:
            self.borderItem.setPen(QtGui.QPen(QtCore.Qt.red))
            self.borderItem.setZValue(100) # above all items
        else:
            self.borderItem.setPen(QtGui.QPen(QtCore.Qt.gray))
            self.borderItem.setZValue(3) # reset to the default z value

    def setMask(self, x_norm, y_norm, w_norm, h_norm):
        # coordinates are in normalized form [0, 1]
        width = self.rect().width()
        height = self.rect().height()
        x = x_norm * width
        y = y_norm * height
        w = w_norm * width
        h = h_norm * height
        self.maskItem.setRect(x, y, w, h)

    def updateTransientItems(self, visibleRect):
        for pid in self.cached_pixmaps:
            pixmap = self.cached_pixmaps[pid]['pixmap']
            raw_dx = self.cached_pixmaps[pid]['dx']
            raw_dy = self.cached_pixmaps[pid]['dy']
            ratio = self.cached_pixmaps[pid]['ratio']
            associated_item = self.cached_pixmaps[pid]['item']
            # 
            x = raw_dx * ratio
            y = raw_dy * ratio
            w = pixmap.width() * ratio
            h = pixmap.height() * ratio
            vrect = QtCore.QRectF(x, y, w, h) # virtual rect
            irect = visibleRect.intersected(vrect) # intersected rect
            # 
            if irect.isEmpty(): # remove associated transient item if unvisible
                if associated_item:
                    associated_item.setParentItem(None)
                    self.cached_pixmaps[pid]['item'] = None
            else:               
                # crop the visible part
                if False:
                    # Qt version, sub-pixel is not supported, heavy jitter effect when drag
                    crop_x = (irect.x() - vrect.x()) / ratio
                    crop_y = (irect.y() - vrect.y()) / ratio
                    crop_w = irect.width() / ratio
                    crop_h = irect.height() / ratio
                    cropped_pixmap = pixmap.copy(crop_x, crop_y, crop_w, crop_h)
                    # resize to current scale
                    resized_pixmap = cropped_pixmap.scaled(math.ceil(irect.width()), math.ceil(irect.height()))
                else:
                    # OpenCV version, support sub-pixel accuracy
                    cvImg = QImg2cvImg(QtGui.QImage(pixmap))
                    # 
                    offset_x = irect.x() - vrect.x()
                    offset_y = irect.y() - vrect.y()
                    M = np.array([[ratio, 0.0, -offset_x], [0.0, ratio, -offset_y], [0.0, 0.0, 1.0]])
                    cvImg = cv2.warpAffine(
                        cvImg, M[:2], (math.ceil(irect.width()), math.ceil(irect.height())),
                        flags=cv2.INTER_NEAREST,
                        borderMode=cv2.BORDER_REPLICATE
                    )
                    # 
                    qImg = cvImg2QImg(cvImg)
                    resized_pixmap = QtGui.QPixmap.fromImage(qImg)

                # debug, draw some mask
                # if True:
                if False:
                    painter = QtGui.QPainter()
                    painter.begin(resized_pixmap)
                    painter.fillRect(resized_pixmap.rect(), QtGui.QColor(255, 0, 0, 100))
                    painter.end()
                    debug("Add transient patch: %d" % pid)

                if associated_item:
                    associated_item.setPixmap(resized_pixmap)
                    associated_item.setOffset(irect.x(), irect.y())
                else:
                    # add to scene
                    item = QtWidgets.QGraphicsPixmapItem(resized_pixmap, parent=self)
                    item.setOffset(irect.x(), irect.y())
                    item.setZValue(1)
                    self.cached_pixmaps[pid]['item'] = item

    def get_roi_patches(self, roi_rect):
        patch_positions = []
        patches = []
        for i in range(self.patch_row_num):
            for j in range(self.patch_col_num):
                pRect = self.patch_rects[i][j]
                interRect = roi_rect.intersected(pRect)
                if not interRect.isEmpty():
                    patches.append(pRect)
                    patch_positions.append([i, j])
        return patch_positions, patches
    
    def get_patch_id(self, i, j, col_num):
        # assert row_num == col_num
        layerBase = col_num * col_num
        offset = i * col_num + j
        return int(layerBase + offset)

    def get_containing_patch_id(self, x, y):
        for i in range(self.patch_row_num):
            for j in range(self.patch_col_num):
                pRect = self.patch_rects[i][j]
                if pRect.contains(x, y):
                    return self.get_patch_id(i, j, self.patch_col_num)
        return None

    def getParentPatchId(self, patch_id):
        # get the parent patch id of patch_id
        if patch_id == 1: # 1 is the root id
            return None
        # get current column number
        col_num = int(math.log2(math.sqrt(patch_id)))
        col_num = int(math.pow(2, col_num))
        # get current offset
        offset = patch_id - (col_num * col_num)
        i = int(offset / col_num)
        j = int(offset % col_num)
        parent_i = int(i / 2)
        parent_j = int(j / 2)
        parent_col_num = int(col_num / 2)
        parent_id = self.get_patch_id(parent_i, parent_j, parent_col_num)
        return parent_id

    def isChildPatch(self, child_patch_id, parent_patch_id):
        # test whether child_patch_id is a child of parent_patch_id
        pid = self.getParentPatchId(child_patch_id)
        while pid is not None and pid >= parent_patch_id:
            if pid == parent_patch_id:
                return True
            pid = self.getParentPatchId(pid)
        return False

    def compute_patch_rects(self, width, height):
        rects = []
        # the width (and also the height) will be divided to [1,2,4,8,16,...] peices

        length = min(width, height)
        if length >= self.patch_basesize:
            splitCnt = int(math.pow(2, int(math.log2(length / self.patch_basesize) + 0.5)))
        else:
            splitCnt = 1

        self.patch_row_num = splitCnt
        self.patch_col_num = splitCnt

        horiStep = int(width / self.patch_col_num)
        vertStep = int(height / self.patch_row_num)
        for i in range(self.patch_row_num):
            tmpRects = []
            for j in range(self.patch_col_num):
                x = j*horiStep
                y = i*vertStep
                w = horiStep
                h = vertStep
                # 
                if x + w > width:
                    w = width - x
                if y + h > height:
                    h = height - y
                tmpRects.append(QtCore.QRectF(x, y, w, h))
            rects.append(tmpRects)

        return rects

    def addPixmap(self, pixmap, dx, dy):
        item = QtWidgets.QGraphicsPixmapItem(pixmap, parent=self)
        item.setOffset(dx, dy)
        item.setZValue(2)

        current_id = self.get_containing_patch_id(dx + pixmap.width() / 2, dy + pixmap.height() / 2)
        # debug("Add current patch id: ", current_id)

        self.current_items[current_id] = {
            "item": item,
            "pixmap": pixmap,
            "dx": dx,
            "dy": dy,
            "ratio": 1.0,
        }

        # remove cached pixmaps which are child of the current
        pidToRemove = []
        for cached_id in self.cached_pixmaps:
            if cached_id == current_id or self.isChildPatch(cached_id, current_id):
                pidToRemove.append(cached_id)
                # debug("Remove cached patch id: ", cached_id)
        for pid in pidToRemove:
            self.removeCachedPixmap(pid)

        # debug
        # if True:
        if False:
            outStr = ("Current items (%d): " % len(self.current_items))
            for pid in self.current_items:
                outStr += ("%d " % pid)
            debug(outStr)

            outStr = ("Cached items (%d): " % len(self.cached_pixmaps))
            for pid in self.cached_pixmaps:
                outStr += ("%d " % pid)
            debug(outStr)
        
    def removeCachedPixmap(self, pid):
        if pid in self.cached_pixmaps:
            associated_item = self.cached_pixmaps[pid]['item']
            if associated_item:
                associated_item.setParentItem(None)
            self.cached_pixmaps.pop(pid)

    def clear(self):
        self.setRect(0, 0, 0, 0)
        self.maskItem.setRect(0, 0, 0, 0)

        self.patch_rects = []
        self.cached_pixmaps = {}
        self.current_items = {}
        cachedItems = self.childItems()
        for item in cachedItems:
            item.setParentItem(None)

# class DocumentFrame(QtWidgets.QFrame):
#     """ This widget is a container of PageWidgets. PageWidget communicates
#         Window through this widget """
#     jumpToRequested = QtCore.pyqtSignal(int, float)
#     copyTextRequested = QtCore.pyqtSignal(int, QtCore.QPoint, QtCore.QPoint)
#     showStatusRequested = QtCore.pyqtSignal(str)
#     pagePositionChanged = QtCore.pyqtSignal(int, int)
#     # 
#     # internal useage

#     # parent is scrollAreaWidgetContents
#     def __init__(self, parent, scrollArea, doc, screen_dpi, threads = 4):
#         super(DocumentFrame, self).__init__(parent)

#         self.setFrameShape(QtWidgets.QFrame.StyledPanel)
#         self.setFrameShadow(QtWidgets.QFrame.Raised)

#         self.scrollArea = scrollArea
#         self.vScrollbar = scrollArea.verticalScrollBar()
#         self.hScrollbar = scrollArea.horizontalScrollBar()

#         self.scrollArea.resizeRequested.connect(self.handleScrollAreaResized)
#         self.scrollArea.verticalScrollBar().setValue(0)
#         self.scrollArea.verticalScrollBar().valueChanged.connect(self.renderCurrentVisiblePages)

#         self.setMouseTracking(True)
#         self.clicked = False

#         self.copy_text_mode = False

#         # 
#         self.doc = doc
#         self.pages_count = self.doc.numPages()
#         #
#         self.pre_empty_page_count = 0
#         self.view_column_count = 1

#         self.layout = QtWidgets.QGridLayout(self)
#         self.layout.setContentsMargins(0, 0, 0, 0) # left, top, right, bottom
#         self.layout.setAlignment(QtCore.Qt.AlignTop | QtCore.Qt.AlignCenter)
#         self.layout.setHorizontalSpacing(0)
#         self.layout.setVerticalSpacing(0)

#         self.setContentsMargins(0, 0, 0, 0)
        
#         self.page_widgets = []
#         for i in range(self.pages_count):
#             for j in range(4):
#                 page = PageWidget(i, self)
#                 self.layout.addWidget(page, i, j, QtCore.Qt.AlignCenter)

#         self.current_page = 0
#         self.current_render_dpi = 0 
#         self.zoom_fit_width = True # default on
#         self.screen_dpi = screen_dpi

#         self.visible_pages = [0]
#         self.rendered_pages = {}
 
#         # create render threads
#         self.num_threads = threads
#         self.render_list = []
#         for i in range(self.num_threads):
#             tmpRender = PdfRender()
#             tmpRender.rendered.connect(self.setRenderedImage)
#             tmpRender.start()
#             self.render_list.append(tmpRender)

#         # set document in other threads
#         for rd in self.render_list:
#             rd.set_document(self.doc)

#         # set default style
#         self.setStyleSheet("border: 0; background-color: gray")

#         self.rearrangePages()

#     def rearrangePages(self):
#         # https://stackoverflow.com/questions/4528347/clear-all-widgets-in-a-layout-in-pyqt/13103617
#         return

#         for item in self.page_widgets:
#             item.setParent(None)
#             item.close()

#         self.page_widgets = [] # clear all pages

#         for i in range(self.pages_count):
#             page = PageWidget(i, self)
#             self.page_widgets.append(page)
#             gridIdx = i + self.pre_empty_page_count
#             row = int (gridIdx / self.view_column_count)
#             col = gridIdx % self.view_column_count
#             self.layout.addWidget(page, row, col, QtCore.Qt.AlignCenter)

#         self.visible_pages = [0]
#         self.rendered_pages = {}

#         if self.zoom_fit_width:
#             self.resizePages(0)
#         else:
#             self.resizePages(self.current_render_dpi)

#     def getVisiblePages(self):
#         visible_pages = []
#         for pg_no in range(self.pages_count):
#             if not self.page_widgets[pg_no].visibleRegion().isEmpty():
#                 visible_pages.append(pg_no)
#         return visible_pages

#     def setRenderedImage(self, page_no, dpi, image):
#         """ takes a QImage and sets pixmap of the specified page
#             when number of rendered pages exceeds a certain number, old page image is
#             deleted to save memory """

#         debug("> Rendering Completed :", page_no)

#         self.page_widgets[page_no].setPageData(page_no, QtGui.QPixmap.fromImage(image), self.doc.page(page_no))
#         self.rendered_pages[page_no] = dpi
#         # # Request to render next page
#         # if self.current_page <= page_no < (self.current_page + self.max_preload - 2):
#         #     if (page_no+2 not in self.rendered_pages) and (page_no+2 <= self.pages_count):
#         #       self.renderRequested.emit(page_no+2, self.pages[page_no+1].dpi)
#         # Replace old rendered pages with blank image

#         if len(self.rendered_pages) > 10:
#             self.visible_pages = self.getVisiblePages()
#             firstKey = next(iter(self.rendered_pages))
#             if firstKey not in self.visible_pages:
#                 self.rendered_pages.pop(firstKey)
#                 # debug("Clear Page :", firstKey)
#                 self.page_widgets[firstKey].clear()

#         # debug("Rendered Pages :", self.rendered_pages)

#     def renderRequest(self, page_no, dpi):
#         tgtWorkerIdx = page_no % self.num_threads
#         self.render_list[tgtWorkerIdx].set_visible_pages(self.visible_pages)    
#         self.render_list[tgtWorkerIdx].render_async(page_no, dpi)

#     def renderCurrentVisiblePages(self):
#         """ Requests to render current page. if it is already rendered, then request
#             to render next unrendered page """

#         return

#         self.visible_pages = self.getVisiblePages()
#         debug("Visible Pages: ", self.visible_pages)

#         if len(self.visible_pages) == 0:
#             return

#         if self.current_page != self.visible_pages[0]:
#             self.current_page = self.visible_pages[0]
#             self.pagePositionChanged.emit(self.current_page, self.pages_count)

#         for page_no in self.visible_pages:
#             if page_no < 0 or page_no >= self.pages_count:
#                 continue
#             if page_no in self.rendered_pages and abs(self.rendered_pages[page_no] - self.current_render_dpi) < 1:
#                 continue
#             self.renderRequest(page_no, self.current_render_dpi)
#             debug("< Rendering Requested :", page_no)

#     def setColumnNumber(self, columnNum):
#         debug("setColumnNumber in DocumentFrame")
#         self.view_column_count = columnNum
#         self.rearrangePages()

#     def setPrecedingEmptyPage(self, emptyCount):
#         debug("setPrecedingEmptyPage %d in DocumentFrame" % emptyCount)
#         self.pre_empty_page_count = emptyCount
#         self.rearrangePages()

#     def setZoom(self, dpi, keep_fit_with=False):
#         """ Gets called when zoom level is changed"""

#         if abs(self.current_render_dpi-dpi) > 0.01:
#             if dpi != 0 and not keep_fit_with:
#                 self.zoom_fit_width = False
#             self.resizePages(dpi)
#             zoomRatio = int(100 * self.current_render_dpi / self.screen_dpi)
#             self.showStatus("Zooming to %d%%" % zoomRatio)

#     def zoomIn(self):
#         dpi = self.current_render_dpi * 1.5
#         if dpi > 300:
#             return
#         self.setZoom(dpi)

#     def zoomOut(self):
#         dpi = self.current_render_dpi * 0.5
#         if dpi < 5:
#             return
#         self.setZoom(dpi)

#     def zoomFitWidth(self):
#         self.zoom_fit_width = True
#         self.setZoom(0)

#     def jumpToCurrentPage(self):
#         """ this is used as a slot, to connect with a timer"""
#         self.jumpToPage(self.current_page)

#     def jumpToPage(self, page_num, top=0.0):
#         """ scrolls to a particular page and position """
#         if page_num < 0: 
#             page_num = 0
#         elif page_num >= self.pages_count: 
#             page_num = self.pages_count - 1
#         if not (0 < top < 1.0) : top = 0
#         self.jumped_from = self.current_page
#         self.current_page = page_num
#         if self.jumped_from != self.current_page:
#             self.pagePositionChanged.emit(self.current_page, self.pages_count)
#         scrollbar_pos = self.page_widgets[page_num].pos().y()
#         scrollbar_pos += top*self.page_widgets[page_num].height()
#         self.docScrollArea1.verticalScrollBar().setValue(scrollbar_pos)

#     def undoJump(self):
#         if self.jumped_from == None: return
#         self.jumpToPage(self.jumped_from)

#     def goNextPage(self):
#         if self.current_page == self.pages_count - 1:
#              return
#         self.jumpToPage(self.current_page + 1)

#     def goPrevPage(self):
#         if self.current_page == 0: 
#             return
#         self.jumpToPage(self.current_page - 1)

#     def gotoPage(self, page_no):
#         self.jumpToPage(page_no)

#     def handleScrollAreaResized(self, width, height):
#         debug("ScrollArea Resized Event received in DocumentFrame: %d, %d" % (width, height))
#         if self.zoom_fit_width:
#             self.setZoom(0, keep_fit_with=True)

#     def compute_fixed_width_dpi(self, pageInchWidth):
#          # remove the space of scroll bar
#         viewPixelWidth = self.scrollArea.width() - self.scrollArea.verticalScrollBar().width() - 4
#         viewPixelWidth -= (self.layout.horizontalSpacing() * (self.view_column_count - 1))
#         viewPixelWidth /= self.view_column_count
#         dpi = viewPixelWidth / pageInchWidth
#         return dpi

#     def resizePages(self, dpi):
#         '''Resize all pages according to zoom level '''

#         # get raw width and height of the current page
#         pg_width = self.doc.page(self.current_page).pageSizeF().width() / 72.0 # width in inch
#         pg_height = self.doc.page(self.current_page).pageSizeF().height() / 72.0

#         new_dpi = dpi
#         if new_dpi == 0: # if fixed width
#             new_dpi = self.compute_fixed_width_dpi(pg_width)

#         # adjust page dimentions
#         for i in range(self.pages_count):
#             self.page_widgets[i].dpi = new_dpi
#             self.page_widgets[i].setFixedSize(pg_width * new_dpi, pg_height * new_dpi)

#         for page_no in self.rendered_pages:
#             self.page_widgets[page_no].updateImage()

#         # if self.zoom_level == 0:
#         #     # adjust scroll postion
#         #     pass

#         # scrolbar_pos = self.pages[self.current_page].pos().y()
#         # self.scrollArea.verticalScrollBar().setValue(scrolbar_pos)

#         self.update() # invoke the paintEvent method using update() explicitly
#         self.current_render_dpi = new_dpi

#         # the page sizes only take effect after next paint event
#         # here use a simple time hacking to wait for its taking effect
#         QtCore.QTimer.singleShot(2000, self.renderCurrentVisiblePages)

#     def showEvent(self, ev):
#         debug('showEvent in DocFrame')
#         # self.resizePages(self.current_render_dpi)

#     def closeEvent(self, ev):
#         """ Save all settings on window close """
#         debug('closeEvent in DocFrame')
#         for rd in self.render_list:
#             rd.stop_async()
#         for rd in self.render_list:
#             rd.wait()
#         return QtWidgets.QFrame.closeEvent(self, ev)

#     # def resizeEvent(self, ev):
#     #     debug('resizeEvent in DocFrame')

#     def wheelEvent(self, ev):
#         debug("wheelEvent in DocFrame")
#         # modifiers = QtWidgets.QApplication.keyboardModifiers()
#         # if modifiers == QtCore.Qt.ControlModifier:
#         #     delta = ev.angleDelta()
#         #     if delta.y() > 0:
#         #         self.zoomIn()
#         #     else:
#         #         self.zoomOut()
#         #     ev.accept() # accept an event in order to stop it from propagating further

#     def mousePressEvent(self, ev):
#         debug("mousePressEvent in DocFrame")
#         self.click_pos = ev.globalPos()
#         self.v_scrollbar_pos = self.vScrollbar.value()
#         self.h_scrollbar_pos = self.hScrollbar.value()
#         self.clicked = True

#     def mouseReleaseEvent(self, ev):
#         debug("mouseReleaseEvent in DocFrame")
#         self.clicked = False
#         self.setCursor(QtCore.Qt.OpenHandCursor)

#     def mouseDoubleClickEvent(self, ev):
#         debug("mouseDoubleClickEvent in DocFrame")
#         self.setZoom(self.screen_dpi)
        
#     def mouseMoveEvent(self, ev):
#         # debug("mouseMoveEvent in DocFrame")
#         if not self.clicked : return
#         self.setCursor(QtCore.Qt.ClosedHandCursor)
#         self.vScrollbar.setValue(self.v_scrollbar_pos + self.click_pos.y() - ev.globalY())
#         self.hScrollbar.setValue(self.h_scrollbar_pos + self.click_pos.x() - ev.globalX())

#     def jumpTo(self, page_num, top):
#         self.jumpToRequested.emit(page_num, top)

#     def enableCopyTextMode(self, enable):
#         self.copy_text_mode = enable

#     def copyText(self, page_num, top_left, bottom_right):
#         self.copyTextRequested.emit(page_num, top_left, bottom_right)

#     def showStatus(self, msg):
#         self.showStatusRequested.emit(msg)


# class PageWidget(QtWidgets.QLabel):
#     """ This widget shows a rendered page """
#     def __init__(self, page_num, frame=None):
#         super(PageWidget, self).__init__(frame)

#         self.manager = frame
#         self.setMouseTracking(True)
#         self.setSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
#         self.link_areas = []
#         self.link_annots = []
#         self.annots_listed, self.copy_text_mode = False, False
#         self.click_point, self.highlight_area = None, None
#         self.page_num = page_num
#         self.image = QtGui.QPixmap()
#         self.mask = PageMaskWidget(self)

#         self.setStyleSheet("border-width: 1px; border-style: solid; border-color: black; background-color: white")
        
#         self.setAlignment(QtCore.Qt.AlignLeft | QtCore.Qt.AlignTop)
#         self.setText("Loading ...")

#         # self.setFixedSize(0, 0)

#     def setPageData(self, page_no, pixmap, page):
#         self.image = pixmap
#         self.updateImage()
#         if self.annots_listed : return
#         annots = page.annotations()
#         for annot in annots:
#             if annot.subType() == Poppler.Annotation.ALink:
#                 x, y = annot.boundary().left()*pixmap.width(), annot.boundary().top()*pixmap.height()
#                 w, h = annot.boundary().width()*pixmap.width()+1, annot.boundary().height()*pixmap.height()+1
#                 self.link_areas.append(QtCore.QRectF(x,y, w, h))
#                 self.link_annots.append(annot)
#         self.annots_listed = True

#     def clear(self):
#         QtWidgets.QLabel.clear(self)
#         self.setText("Loading ...")
#         self.image = QtGui.QPixmap()

#     def mouseMoveEvent(self, ev):
#         # Draw rectangle when mouse is clicked and dragged in copy text mode.
#         if self.manager.copy_text_mode:
#             if self.click_point:
#                 pm = self.pm.copy()
#                 painter = QtGui.QPainter()
#                 painter.begin(pm)
#                 painter.drawRect(QtCore.QRect(self.click_point, ev.pos()))
#                 painter.end()
#                 self.setPixmap(pm)
#             return

#         # Change cursor if cursor is over link annotation
#         for i, area in enumerate(self.link_areas):
#             if area.contains(ev.pos()):
#                 linkDest = self.link_annots[i].linkDestination()
#                 if not linkDest : continue
#                 # For jump to page link
#                 if linkDest.linkType() == Poppler.Link.Goto:
#                     p = linkDest.destination().pageNumber()
#                     self.manager.showStatus("Jump To Page : %i" % p)
#                     self.setCursor(QtCore.Qt.PointingHandCursor)
#                 # For URL link
#                 elif linkDest.linkType() == Poppler.Link.Browse:
#                     self.manager.showStatus("URL : %s" % linkDest.url())
#                     self.setCursor(QtCore.Qt.PointingHandCursor)
#                 return
#         # self.manager.showStatus("")
#         self.unsetCursor()
#         ev.ignore()         # pass to underlying frame if not over link or copy text mode

#     def mousePressEvent(self, ev):
#         # In text copy mode
#         if self.manager.copy_text_mode:
#             self.click_point = ev.pos()
#             self.pm = self.pixmap().copy()
#             return
#         # In normal mode
#         for i, area in enumerate(self.link_areas):
#             if not area.contains(ev.pos()): continue
#             link_dest = self.link_annots[i].linkDestination()
#             if not link_dest : continue
#             # For jump to page link
#             if link_dest.linkType() == Poppler.Link.Goto:
#                 page_num = link_dest.destination().pageNumber()
#                 top = 0.0
#                 if link_dest.destination().isChangeTop():
#                     top = link_dest.destination().top()
#                 self.manager.jumpTo(page_num, top)
#             # For URL link
#             elif link_dest.linkType() == Poppler.Link.Browse:
#                 url = link_dest.url()
#                 if url.startswith("http"):
#                     confirm = QMessageBox.question(self, "Open Url in Browser",
#                         "Do you want to open browser to open...\n%s" %url, QMessageBox.Yes|QMessageBox.Cancel)
#                     if confirm == QMessageBox.Yes:
#                         Popen(["x-www-browser", url])
#             return
#         ev.ignore()

#     def mouseReleaseEvent(self, ev):
#         if self.manager.copy_text_mode:
#             self.manager.copyText(self.page_num, self.click_point, ev.pos())
#             self.setPixmap(self.pm)
#             self.click_point = None
#             self.pm = None
#             return
#         ev.ignore()

#     def updateImage(self):
#         """ repaint page widget, and draw highlight areas """
#         if self.highlight_area:
#             img = self.image.copy()
#             painter = QPainter(img)
#             zoom = self.dpi / 72.0
#             for area in self.highlight_area:
#                 box = QtCore.QRectF(area.left()*zoom, area.top()*zoom,
#                                     area.width()*zoom, area.height()*zoom)
#                 painter.fillRect(box, QColor(0,255,0, 127))
#             painter.end()
#             self.setPixmap(img.scaled(self.width(), self.height(), transformMode=QtCore.Qt.SmoothTransformation))
#             # self.setPixmap(img)
#         else:
#             # self.setPixmap(self.image)
#             self.setPixmap(self.image.scaled(self.width(), self.height(), transformMode=QtCore.Qt.SmoothTransformation))

# class PageMaskWidget(QtWidgets.QLabel):
#     """ This widget shows a rendered page """
#     def __init__(self, parent):
#         super(PageMaskWidget, self).__init__(parent)

#         self.setMouseTracking(True)
#         self.setSizePolicy(0,0)
#         self.setStyleSheet("border-width: 1px; border-style: solid; border-color: yellow; background-color: rgba(0, 0, 0, 100)")
#         self.setGeometry(0,0,0,0) # the default is unvisible


 
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from utils import debug

class PageNoLineEdit(QtWidgets.QLineEdit):

    def __init__(self, parent):
        super(PageNoLineEdit, self).__init__(parent)
        self.parent = parent

        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setFixedWidth(50)
        self.setMaxLength(4)

        validator = QtGui.QIntValidator(1, 9999, self)
        self.setValidator(validator)

    def mousePressEvent(self, event):
        self.selectAll()

        # print('forwarding to the main window')
        self.parent.mousePressEvent(event)

class FindLineEdit(QtWidgets.QLineEdit):

    def __init__(self, parent):
        super(FindLineEdit, self).__init__(parent)
        self.parent = parent

        self.setAlignment(QtCore.Qt.AlignCenter)
        self.setFixedWidth(50)
        self.setMaxLength(4)

        validator = QtGui.QIntValidator(1, 9999, self)
        self.setValidator(validator)

    def mousePressEvent(self, event):
        self.selectAll()

        # print('forwarding to the main window')
        self.parent.mousePressEvent(event)


import os

from resources.ui_document import Ui_doc_view
from PyQt5 import QtCore
from PyQt5 import QtGui
from PyQt5 import QtWidgets

from popplerqt5 import Poppler

import math

from utils import debug

from pdfworker import PdfReader
from annotations import AnnotationFrame, AnnotationItemWidget

from __init__ import __version__

class DocumentView(QtWidgets.QWidget, Ui_doc_view):
    readAnnotationRequested = QtCore.pyqtSignal(Poppler.Document)
    readOutlineRequested = QtCore.pyqtSignal(Poppler.Document, QtGui.QStandardItemModel)
    findTextRequested = QtCore.pyqtSignal(str, int, bool)
    pagePositionChanged = QtCore.pyqtSignal(int, int)
    showStatusRequested = QtCore.pyqtSignal(str)

    def __init__(self, parent, filename, screen_dpi):
        super(DocumentView, self).__init__(parent) # Call the inherited classes __init__ method
        self.setupUi(self)

        self.filename = filename
        self.screen_dpi = screen_dpi

        self.doc = None
        self.setWindowTitle(os.path.basename(self.filename)+ " - kuafu " + __version__)

        # hide some parts in splitters
        self.splitter_main.setSizes([1,5,1])
        self.splitter_doc.setSizes([1,0])

        self.splitter_main.setCollapsible(0, False) # make the thumb view uncollapsible
        self.splitter_main.setCollapsible(2, False) # make the annotationview uncollapsible
        self.splitter_doc.setCollapsible(0, False) # make the first document view uncollapsible

        # Create separate thread for reading pdf metainformations
        self.pdf_reader_thread = QtCore.QThread(self)
        self.pdf_reader_obj = PdfReader()
        self.pdf_reader_obj.moveToThread(self.pdf_reader_thread) # this must be moved before connecting signals
        # self.renderRequested.connect(self.pdf_reader_obj.render)
        self.readAnnotationRequested.connect(self.pdf_reader_obj.readAnnotation)
        self.readOutlineRequested.connect(self.pdf_reader_obj.readOutline)
        # self.findTextRequested.connect(self.pdf_reader_obj.findText)
        self.pdf_reader_obj.annotationFound.connect(self.onAnnotaionFound)
        self.pdf_reader_obj.outlineFound.connect(self.onOutlineFound)
        # self.pdf_reader_obj.textFound.connect(self.onTextFound)
        self.pdf_reader_thread.start()

        # self.doc = Poppler.Document.load(filename)
        # if not self.doc : return

        # if self.doc.isLocked() :
        #     password = QInputDialog.getText(self, 'This PDF is locked', 'Enter Password :', 2)[0]
        #     if password == '' : sys.exit(1)
        #     self.doc.unlock(password.encode(), password.encode())
        # if not self.first_document:
        #     self.removeOldDoc()

        password = ""
        self.doc = Poppler.Document.load(filename, password.encode(), password.encode())
        self.doc.setRenderHint(
            Poppler.Document.TextAntialiasing
            | Poppler.Document.TextHinting
            | Poppler.Document.Antialiasing
            # | Poppler.Document.ThinLineSolid
            )

        self.pages_count = self.doc.numPages()

        self.outline_model = QtGui.QStandardItemModel(self)

        self.readOutlineRequested.emit(self.doc, self.outline_model)
        self.readAnnotationRequested.emit(self.doc)
        
        # self.zoomLevelCombo.setCurrentIndex()

        # if collapseUser(self.filename) in self.history_filenames:
        #     self.current_page = int(self.history_page_no[self.history_filenames.index(collapseUser(self.filename))])
        # self.current_page = min(self.current_page, self.pages_count)
        self.current_page = 0

        self.graphicsView_doc1.setDocument(self.doc, self.screen_dpi, render_num=4)
        self.graphicsView_doc2.setDocument(self.doc, self.screen_dpi, render_num=4)
        self.graphicsView_thumbs.setDocument(self.doc, self.screen_dpi, render_num=4)
        
        # # Add document 1 widgets
        # self.doc_frame1 = DocumentFrame(self.docScrollAreaContents1, self.docScrollArea1, self.doc, self.screen_dpi, threads=4)
        # self.docScrollAreaContents1_Layout.addWidget(self.doc_frame1)
        # self.doc_frame1.pagePositionChanged.connect(self.handlePagePositionChanged)

        self.graphicsView_doc1.pagePositionChanged.connect(self.onPagePositionChanged)
        # self.docScrollArea1.verticalScrollBar().valueChanged.connect(self.onDocScrollAreaChanged)

        # # Add document 2 widgets
        # self.view_doc1 = DocumentFrame(self.docScrollAreaContents2, self.docScrollArea2, self.doc, self.screen_dpi, threads=4)
        # self.docScrollAreaContents2_Layout.addWidget(self.doc_frame2)
        # self.doc_frame2.pagePositionChanged.connect(self.handlePagePositionChanged)

        # # Add thumbs widgets
        # self.thumbs_frame = DocumentFrame(self.thumbsScrollAreaContents, self.thumbsScrollArea, self.doc, self.screen_dpi, threads=4)
        # self.thumbScrollAreaContents_Layout.addWidget(self.thumbs_frame)

        # Add annotation widgets
        self.annotation_frame = AnnotationFrame(self.annotScrollAreaContents, self.annotScrollArea)
        self.annotScrollAreaContents_Layout.addWidget(self.annotation_frame)

        # self.thumbsScrollArea.resizeRequested.connect(self.thumbs_frame.handleScrollAreaResized)
        # self.thumbsScrollArea.verticalScrollBar().setValue(0)
        # self.thumbsScrollArea.verticalScrollBar().valueChanged.connect(self.thumbs_frame.renderCurrentVisiblePages)

        # self.doc_frame1.showStatusRequested.connect(self.showStatus)
        # self.doc_frame2.showStatusRequested.connect(self.showStatus)
        # self.thumbs_frame.showStatusRequested.connect(self.showStatus)

        # Set style
        # self.thumbs_frame.setStyleSheet("border: 0; background-color: white")

        # self.annotationItemModel = QtGui.QStandardItemModel()
        # self.annotationView.setModel(self.annotationItemModel)
        # self.annotationView.resizeRequested.connect(self.onAnnotaionViewResized)
 
        # self.annotationListView.setAlternatingRowColors(True)
        # self.annotationListView.clicked.connect(self.onOutlineClick)

        # self.doc_frame1.jumpToRequested.connect(self.jumpToPage)
        # self.doc_frame1.copyTextRequested.connect(self.copyText)
        # self.doc_frame1.showStatusRequested.connect(self.showStatus)

        # self.pageNoLabel.setText('<b>%i/%i</b>' % (self.current_page, self.pages_count) )
        # self.gotoPageValidator.setTop(self.pages_count)

        # if self.current_page != 0 :
        #     QtCore.QTimer.singleShot(150+self.pages_count//3, self.jumpToCurrentPage)

    def showStatus(self, msg):
        self.showStatusRequested.emit(msg)

    def onAnnotaionFound(self, annotList):
        assert(len(annotList) == self.pages_count)
        for pg_no in range(self.pages_count):
            for annot in annotList[pg_no]:
                image = None
                author = annot['author']
                boundary = annot['boundary']
                contents = annot['contents']
                modificationDate = annot['modificationDate']
                style = annot['style']
                if annot['type'] == 'highlight':
                    text = annot['text']
                    self.annotation_frame.addItem(style.color(), author, modificationDate, text, contents)
                elif annot['type'] == 'geom':
                    image = annot['image']
                    self.annotation_frame.addItem(style.color(), author, modificationDate, image, contents)

    def onOutlineFound(self, itemModel):
        debug("outline found")

        self.treeView_outline.setModel(itemModel)
        if itemModel.invisibleRootItem().rowCount() < 4:
            self.treeView_outline.expandToDepth(0)
        self.treeView_outline.setHeaderHidden(True)
        self.treeView_outline.header().setSectionResizeMode(0, 1)
        self.treeView_outline.header().setSectionResizeMode(1, 3)
        self.treeView_outline.header().setStretchLastSection(False)
        self.treeView_outline.expandAll()
        self.treeView_outline.setAlternatingRowColors(True)
        self.treeView_outline.clicked.connect(self.onOutlineClick)

    def onPagePositionChanged(self, current_page, pages_count):
        return
        
        target_view = self.graphicsView_doc1
        thumb_view = self.graphicsView_thumbs

        for i in range(target_view.page_counts-1, -1, -1):

            docItem = target_view.page_items[i]
            thumbItem = thumb_view.page_items[i]
            thumbMaskItem = thumb_view.page_mask_items[i]

            visRegion = target_view.visible_regions[i]

            if visRegion.isEmpty():
                thumbMaskItem.setRect(0,0,0,0)
            else:
                thumbItem.ensureVisible()

                ratio = thumbItem.boundingRect().width() / docItem.boundingRect().width()
                x = round(visRegion.x() * ratio)
                y = round(visRegion.y() * ratio)
                width = round(visRegion.width() * ratio) - 1
                height = round(visRegion.height() * ratio) - 1
                thumbMaskItem.setRect(x,y,width,height)

        self.pagePositionChanged.emit(current_page, pages_count)

    def docInfo(self):
        info_keys = list(self.doc.infoKeys())
        values = [self.doc.info(key) for key in info_keys]
        page_size = self.doc.page(self.current_page).pageSizeF()
        page_size = "%s x %s pts"%(page_size.width(), page_size.height())
        info_keys += ['Embedded FIles', 'Page Size']
        values += [str(self.doc.hasEmbeddedFiles()), page_size]
        dialog = DocInfoDialog(self)
        dialog.setInfo(info_keys, values)
        dialog.exec_()

    def setColumnNumber(self, columNum):
        self.graphicsView_doc1.setColumnNumber(columNum)
        self.graphicsView_thumbs.setColumnNumber(columNum)

    def setPrecedingEmptyPage(self, emptyCount):
        self.graphicsView_doc1.setPrecedingEmptyPage(emptyCount)
        self.graphicsView_thumbs.setPrecedingEmptyPage(emptyCount)

    def zoomIn(self):
        self.graphicsView_doc1.zoomIn()

    def zoomOut(self):
        self.graphicsView_doc1.zoomOut()

    def zoomFitWidth(self):
        self.graphicsView_doc1.zoomFitWidth()

    def toggleFindMode(self, enable):
        if enable:
          self.findTextEdit.setText('')
          self.findTextEdit.setFocus()
          self.search_text = ''
          self.search_result_page = 0
        elif self.search_result_page != 0:
          self.pages[self.search_result_page].highlight_area = None
          self.pages[self.search_result_page].updateImage()

    def findNext(self):
        """ search text in current page and next pages """
        text = self.findTextEdit.text()
        if text == "" : return
        # search from current page when text changed
        if self.search_text != text or self.search_result_page == 0:
            search_from_page = self.current_page
        else:
            search_from_page = self.search_result_page + 1
        self.findTextRequested.emit(text, search_from_page, False)
        if self.search_result_page != 0:     # clear previous highlights
            self.pages[self.search_result_page].highlight_area = None
            self.pages[self.search_result_page].updateImage()
            self.search_result_page = 0
        self.search_text = text

    def findBack(self):
        """ search text in pages before current page """
        text = self.findTextEdit.text()
        if text == "" : return
        if self.search_text != text or self.search_result_page == 0:
            search_from_page = self.current_page
        else:
            search_from_page = self.search_result_page - 1
        self.findTextRequested.emit(text, search_from_page, True)
        if self.search_result_page != 0:
            self.pages[self.search_result_page].highlight_area = None
            self.pages[self.search_result_page].updateImage()
            self.search_result_page = 0
        self.search_text = text

    def onTextFound(self, page_no, areas):
        self.pages[page_no].highlight_area = areas
        self.search_result_page = page_no
        if self.pages[page_no].pixmap():
            self.pages[page_no].updateImage()
        first_result_pos = areas[0].y()/self.doc.page(page_no).pageSize().height()
        self.jumpToPage(page_no, first_result_pos)

    def toggleCopyText(self, checked):
        self.doc_frame1.enableCopyTextMode(checked)

    def copyText(self, page_no, top_left, bottom_right):
        zoom = self.pages[page_no].height()/self.doc.page(page_no).pageSize().height()
        # Copy text to clipboard
        text = self.doc.page(page_no).text(QtCore.QRectF(top_left/zoom, bottom_right/zoom))
        QApplication.clipboard().setText(text)
        self.copyTextAction.setChecked(False)
        self.toggleCopyText(False)

    def onOutlineClick(self, m_index):
        page_num = self.outlineTreeView.model().data(m_index, QtCore.Qt.UserRole+1)
        top = self.outlineTreeView.model().data(m_index, QtCore.Qt.UserRole+2)
        if not page_num: return
        self.jumpToPage(page_num, top)

    def closeEvent(self, ev):
        """ Save all settings on window close """
        debug('closeEvent in DocView')

        # wait for the reader thread to exit
        loop = QtCore.QEventLoop()
        self.pdf_reader_thread.finished.connect(loop.quit)
        self.pdf_reader_thread.quit()
        loop.exec_()
        debug('pdf_reader_thread exit')

        self.graphicsView_doc1.close()
        self.graphicsView_doc2.close()
        self.graphicsView_thumbs.close()
